
(* ESERCIZIO: SOLUZIONE *)

open List;;
exception Undefined;;

module type RELAZIONE =
 sig type ('a,'b) relazione 
  val relazioneC: unit -> ('a,'b) relazione 
  val addPair: ('a,'b) relazione -> 'a -> 'b -> ('a,'b) relazione 
  val removePair: ('a,'b) relazione -> 'a -> 'b -> ('a,'b) relazione
  val isIn: ('a,'b) relazione -> 'a -> 'b -> bool
  val getUno: ('a,'b) relazione -> 'b -> 'a list
  val getDue: ('a,'b) relazione -> 'a -> 'b list
(* additional *)
  val elements: ('a,'b) relazione -> ('a * 'b) list
  val toString: ('a,'b) relazione -> ('a -> string) -> ('b -> string) -> string
end;;
 
(* complerare con AF, I, additionals e caso di uso 
module Relazione =
 (struct
  type ('a,'b) relazione = ('a * 'b) list
  let relazioneC = fun () -> []
  let addPair r x y = (x,y)::r 
  let removePair r x y = 
     let g = fun u -> if u!=(x,y) then [u] else []
     in fold_right(append)(map g r)[]
  let isIn r x y = fold_right (||) (map ((=)(x,y)) r) false 
  let isDue(r,y) = fold_right (||) (map(fun u -> snd(u)=y) r) false (* inutile, perchè... *)
  let getUno r y = let g = fun u ->if(snd(u)=y)then[fst(u)]else[]
                     in fold_right(append)(map g r) []
  let getDue r x = let g = fun u ->if(fst(u)=x)then[snd(u)]else[]
                     in fold_right (append) (map g r) []
   
end:RELAZIONE);;
*)

(*oppure*)

(* complerare con AF, I, additionals e caso di uso
module Relazione1 =
 (struct
  type ('a,'b) relazione = ('a * 'b) -> bool
  let relazioneC = fun () -> fun u -> false
  let addPair r x y = fun u -> if u=(x,y) then true else (r u)
  let removePair r x y = fun u -> if u=(x,y) then false else (r u)
  let isIn r x y = r(x,y)
  let getUno r y = raise Undefined (*impossibile senza dominio delle coppie*) 
  let getDue r x = raise Undefined (*semidecidibile se dominio coppie numerabile*)
end:RELAZIONE);;
*)

(* completare con AF ed I *)
module Relazione2 =
 (struct
  type ('a,'b) relazione = ((('a * 'b) list) * (('a * 'b) -> bool))
  let relazioneC = fun () -> ([],fun u -> false)
  let addPair r x y = ((x,y)::fst(r),
                       fun u -> if u=(x,y) then true else (snd(r)u))
  let removePair r x y = (fst(r),
                          fun u -> if u=(x,y) then false else (snd(r)u))
  let isIn r x y = (snd r)(x,y)
  let getUno r y = let g = fun u -> if (snd(u)=y) then [u] else []
                   in let drop = fold_right(append)(map g (fst(r))) []
                      in let boolPairs = combine (map(snd(r))drop) drop
                         in let h = fun bp -> if fst(bp) then [fst(snd(bp))] else []
                            in fold_right (append)(map h boolPairs) []                         
  let getDue r x = [] (* ovviamente errata: da scrivere *)
  let elements r = let cons1 x l = match x with [u] -> (u::l) | _ -> l and
  				       g u = if (snd r u) then [u] else [] in
  				   fold_right (cons1)(map g (fst r)) []  
  let toString r aToString bToString = 
  				   let pairString (x,y) = "("^(aToString x) ^ "," ^ (bToString y) ^ ")" in
  				   let myList = elements r in
  				   let myprefix = match myList with u::l -> "{"^(pairString u) |_ -> "{" in
  				   let myfun = fun x l -> ("," ^ pairString x) ^ l in
  				   myprefix ^ (fold_right (myfun) (tl myList) "") ^ "}"
end:RELAZIONE);;


(* Una soluzione con un intero che conta le coppie contenute: 
   Completare con AF, I, additionals e caso di uso 
module Relazione3 =
 (struct
  type ('a,'b) relazione = (int * (('a * 'b) list))
  let relazioneC = 
  let addPair r x y = 
  let isIn r x y =
  let removePair r x y = 
  let getUno r y = 
  let getDue r x =  
end:RELAZIONE);;
*)


(* Una soluzione che utilizza tipi concreti:
   con AF, I I, additionals e caso di uso
module Relazione4 =
 (struct 
  type ('a,'b) relazione = Cons of 'a * 'b * ('a,'b) relazione | Empty
  let relazioneC = fun () -> Empty
  let rec removePair r x y = match r with
      Empty -> Empty
     |Cons(ux,uy,rr) -> if (ux=x) && (uy =y)
                        then rr  (* no duplications *)
                        else Cons(ux,uy,(removePair rr x y))
  let rec isIn r x y = match r with
      Empty -> false
     |Cons(ux,uy,rr) -> ((ux,uy)=(x,y)) || (isIn rr x y) 
  let addPair r x y = if (isIn r x y) then r else Cons(x,y,r)
  let rec getUno r y = match r with
      Empty -> []
     |Cons(ux,uy,rr) -> if (uy =y) then ux::(getUno rr y) else (getUno rr y)
  let getDue r x = [] (* ovviamente errata: da riscrivere *)
end:RELAZIONE);;
*)


(* Caso di uso *)
open Relazione2;;
let valore2 = relazioneC();;
let valoreA2 = addPair valore2 3 5;;
getUno valoreA2 5;;
let valoreBA2 = removePair valoreA2 3 7;;
getUno valoreBA2 5;;
elements valore2;;
elements valoreA2;;
elements valoreBA2;;
let valoreBA2 = addPair valoreA2 3 7;;
let valoreCA2 = addPair valoreBA2 3 9;;
toString valoreCA2 (string_of_int) (string_of_int);;


